#include "lion/icc/ICC.h"
#include "lion/icc/transport/ICC_Stream.h"
#include "lion/log/log.h"
#include "lion/rtos/RTOS_Sleep.h"
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <termios.h>
#include <unistd.h>

LOG_MODULE_DECLARE(app, LOG_LEVEL_DBG);

#define CONFIG_APP_szMODEM     "/dev/ttymxc1"
#define CONFIG_APP_u32BAUDRATE B115200

/* Linux uses an non-configurable 4096 byte buffer for UART. But the available capacity of this buffer is inconsistent.
 * Sometimes it's 3972, sometimes 3969, sometimes 4095. So an artificial limit is placed when using bolAll to quarantee
 * expected operation.
 * We also chunk up the writes into smaller bits so we can maintain write speed. The backend doesn't clear the buffer
 * until each write "block" is fully transmitted. So if a write of 256 bytes is made, the buffer will show 256 bytes
 * used until all 256 bytes are transmitted. Chunking up the writes allows us to keep pushing data into the buffer
 * without the transmitter halting. */
#define CONFIG_APP_sizWriteBuffer 512
#define CONFIG_APP_sizWriteChunk  128

ICC_Context_t APP_strICC;

static pthread_t            app_icc_strThread;
static ICC_Stream_Context_t app_strICCStream;
static volatile bool        app_icc_bolKeepRunning = true;
static int                  app_icc_fdModem;

size_t app_icc_Write(const uint8_t *ptrData, size_t sizLength, bool bolAll, void *ptrUserData)
{
    StackScop int int_Queued;
    StackScop int int_RC = ioctl(app_icc_fdModem, TIOCOUTQ, &int_Queued);
    if (int_RC != 0)
    {
        LOG_ERR("ioctl Error: \"%s\"", strerror(errno));
        return 0;
    }
    StackScop size_t siz_Space = CONFIG_APP_sizWriteBuffer - icc_MIN(int_Queued, CONFIG_APP_sizWriteBuffer);
    if (bolAll)
    {
        if (sizLength > siz_Space)
        {
            return 0;
        }
    }
    ssize_t sizResult =
        write(app_icc_fdModem, ptrData, icc_MIN(icc_MIN(sizLength, siz_Space), CONFIG_APP_sizWriteChunk));
    if (sizResult < 0)
    {
        sizResult = 0;
        switch (errno)
        {
            case EAGAIN:
                /* Do Nothing */
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                break;
        }
    }
    return sizResult;
}

size_t app_icc_Read(uint8_t *ptrData, size_t sizMaxLength, bool bolAll, void *ptrUserData)
{
    if (bolAll)
    {
        StackScop int int_Available;
        StackScop int int_RC = ioctl(app_icc_fdModem, FIONREAD, &int_Available);
        if (int_RC != 0)
        {
            LOG_ERR("ioctl Error: \"%s\"", strerror(errno));
            return 0;
        }
        if (int_Available < sizMaxLength)
        {
            return 0;
        }
    }
    ssize_t sizResult = read(app_icc_fdModem, ptrData, sizMaxLength);
    if (sizResult < 0)
    {
        sizResult = 0;
        switch (errno)
        {
            case EAGAIN:
                /* Do Nothing */
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                break;
        }
    }
    return sizResult;
}

void *app_icc_Entry(void *arg)
{
    (void)arg;

    while (app_icc_bolKeepRunning)
    {
        ICC_Process(&APP_strICC);
        RTOS_Sleep(RTOS_MSEC(10));
    }

    return NULL;
}

void APP_ICC_Init()
{
    ICC_Init(&APP_strICC);
    ICC_Stream_Init(&app_strICCStream, app_icc_Write, app_icc_Read, &app_strICCStream);
    ICC_AddTransport(&APP_strICC, &app_strICCStream.transport);
};

void APP_ICC_Start()
{
    const char *szModem = CONFIG_APP_szMODEM;
    app_icc_fdModem     = open(szModem, O_RDWR | O_NOCTTY | O_NONBLOCK);
    if (app_icc_fdModem < 0)
    {
        const char *szError = strerror(errno);
        LOG_ERR("Unable to open \"%s\" (%s)", szModem, szError ? szError : "None");
        exit(-1);
    }

    /* Setup Taken from pySerial */
    struct termios strOldTIO;
    struct termios strNewTIO;
    int            rc = tcgetattr(app_icc_fdModem, &strOldTIO); /* save current port settings */
    printf("tcgetattr %d\n", rc);
    strNewTIO = strOldTIO;

    /* set up raw mode / no echo / binary */
    strNewTIO.c_cflag |= CLOCAL | CREAD;
    strNewTIO.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHONL | ISIG | IEXTEN | ECHOCTL | ECHOKE);
    strNewTIO.c_oflag &= ~(OPOST | ONLCR | OCRNL);
    strNewTIO.c_iflag &= ~(INLCR | IGNCR | ICRNL | IGNBRK | IUCLC | PARMRK);

    rc = cfsetospeed(&strNewTIO, CONFIG_APP_u32BAUDRATE);
    rc = cfsetispeed(&strNewTIO, CONFIG_APP_u32BAUDRATE);

    /* Setup character length */
    strNewTIO.c_cflag = (strNewTIO.c_cflag & ~CSIZE) | CS8;

    /* Setup stop bits */
    strNewTIO.c_cflag &= ~(CSTOPB); // One
    // strNewTIO.c_cflag |= CSTOPB; // Two

    /* Setup parity */
    strNewTIO.c_iflag &= ~(INPCK | ISTRIP);
    strNewTIO.c_cflag &= ~(PARENB | PARODD | CMSPAR); // None
    // strNewTIO.c_cflag = (strNewTIO.c_cflag & ~(PARODD | CMSPAR)) | (PARENB); // Even
    // strNewTIO.c_cflag = (strNewTIO.c_cflag & ~(CMSPAR)) | (PARENB | PARODD); // Odd

    strNewTIO.c_iflag &= ~(IXON | IXOFF | IXANY); // shut off xon/xoff
    strNewTIO.c_cflag &= ~CRTSCTS;

    strNewTIO.c_cc[VTIME] = 1; /* inter−character timer unused */
    strNewTIO.c_cc[VMIN]  = 0; /* no blocking reads */

    rc = tcflush(app_icc_fdModem, TCIFLUSH);
    rc = tcflush(app_icc_fdModem, TCOFLUSH);
    rc = tcsetattr(app_icc_fdModem, TCSANOW, &strNewTIO);

    pthread_create(&app_icc_strThread, NULL, app_icc_Entry, NULL);
};

void APP_ICC_Stop()
{
    app_icc_bolKeepRunning = false;
    pthread_join(app_icc_strThread, NULL);
    close(app_icc_fdModem);
};