#include "server.h"
#include "lion/icc/ICC.h"
#include "lion/icc/transport/ICC_Stream.h"
#include "lion/log/log.h"
#include "lion/rtos/RTOS_Sleep.h"
#include "lion/utils/dlist.h"
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

LOG_MODULE_DECLARE(server, LOG_LEVEL_INF);

struct icc_server_connection
{
    sys_dnode_t          node;
    pthread_t            thread;
    int                  socket_id;
    ICC_Stream_Context_t strStream;
    volatile bool bolConnected;
};

extern ICC_Context_t APP_strICC;

static volatile bool   icc_server_keepRunning = true;
static pthread_t       icc_server_host_thread;
static pthread_t       icc_server_strICCThread;
static int             icc_server_socket_id;
static sys_dlist_t     icc_server_connections;
static pthread_mutex_t icc_server_mutex;

size_t icc_server_ICCWrite(const uint8_t *ptrData, size_t sizLength, bool bolAll, void *ptrUserData)
{
    struct icc_server_connection *connection = ptrUserData;
    ssize_t                       sizResult  = send(connection->socket_id, ptrData, sizLength, MSG_DONTWAIT | MSG_NOSIGNAL);
    if (sizResult < 0) {
        sizResult = 0;
        switch(errno) {
            case EAGAIN:
                /* Do Nothing */
                break;
            case EPIPE:
                connection->bolConnected = false;
                break;
            default:
                LOG_ERR("Unhandled Send Error: \"%s\"", strerror(errno));
                break;
        }
    }
    return sizResult;
}

size_t icc_server_ICCRead(uint8_t *ptrData, size_t sizMaxLength, bool bolAll, void *ptrUserData)
{
    struct icc_server_connection *connection = ptrUserData;
    ssize_t sizResult = recv(connection->socket_id, ptrData, sizMaxLength, MSG_DONTWAIT | (bolAll ? MSG_WAITALL : 0));
    if (sizResult == 0) {
        connection->bolConnected = false;
    } else if (sizResult < 0) {
        sizResult = 0;
        switch(errno) {
            case EAGAIN:
                /* Do Nothing */
                break;
            case EPIPE:
                connection->bolConnected = false;
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                break;
        }
    }
    return sizResult;
}

void *icc_server_connection_entry(void *arg)
{
    struct icc_server_connection *connection = arg;
    ssize_t                       rc;
    uint8_t                       arrBuffer[4096];                        

    LOG_INF("Connected on %d, Transport %p", connection->socket_id, &connection->strStream);
    connection->bolConnected = true;
    ICC_Stream_Init(&connection->strStream, icc_server_ICCWrite, icc_server_ICCRead, connection);
    ICC_AddTransport(&APP_strICC, &connection->strStream.transport);

    while (icc_server_keepRunning && connection->bolConnected)
    {
        RTOS_Sleep(RTOS_SECONDS(1));
    }

    ICC_RemoveTransport(&APP_strICC, &connection->strStream.transport);

    pthread_mutex_lock(&icc_server_mutex);
    sys_dlist_remove(&connection->node);
    pthread_mutex_unlock(&icc_server_mutex);
    LOG_INF("Terminated on %d, Transport %p", connection->socket_id, &connection->strStream);
    free(connection);

    return NULL;
}

void *icc_server_host_entry(void *arg)
{
    LOG_INF("Startup...");

    struct sockaddr_un socket_addr;

    if ((icc_server_socket_id = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
    {
        exit(-1); // TODO: Properly Throw Error
    }

    memset(&socket_addr, 0, sizeof(socket_addr));
    socket_addr.sun_family = AF_UNIX;
    strcpy(socket_addr.sun_path, "/var/run/lion-icc.sock");

    unlink("/var/run/lion-icc.sock"); // TODO: Properly Throw Error
    if (bind(icc_server_socket_id, (const struct sockaddr *)&socket_addr, sizeof(socket_addr)) < 0)
    {
        LOG_ERR("Unable to bind! \"%s\"", strerror(errno));
        close(icc_server_socket_id);
        exit(-1); // TODO: Properly Throw Error
    }

    if (listen(icc_server_socket_id, 4) < 0)
    {
        LOG_ERR("Unable to listen! \"%s\"", strerror(errno));
        close(icc_server_socket_id);
        exit(-1); // TODO: Properly Throw Error
    }

    while (icc_server_keepRunning)
    {
        struct sockaddr_un connection_addr;
        socklen_t          connection_addrlen = sizeof(connection_addr);
        int connection_id = accept(icc_server_socket_id, (struct sockaddr *)&connection_addr, &connection_addrlen);
        if (connection_id >= 0)
        {
            struct icc_server_connection *connection = malloc(sizeof(struct icc_server_connection));
            connection->socket_id                    = connection_id;
            pthread_mutex_lock(&icc_server_mutex);
            sys_dlist_append(&icc_server_connections, &connection->node);
            pthread_mutex_unlock(&icc_server_mutex);
            pthread_create(&connection->thread, NULL, icc_server_connection_entry, connection);
        }
    }

    close(icc_server_socket_id);
    unlink("/var/run/lion-icc.sock"); // TODO: Properly Throw Error

    LOG_INF("Terminating...");
    return NULL;
}

void icc_server_init(struct icc_server_config *config)
{
    pthread_mutex_init(&icc_server_mutex, NULL);
    sys_dlist_init(&icc_server_connections);
    sigaction(SIGPIPE, &(struct sigaction){SIG_IGN}, NULL); /* Ignore Broken Pipe Signal */
}

void icc_server_start()
{
    pthread_create(&icc_server_host_thread, NULL, icc_server_host_entry, NULL);
}

void icc_server_stop()
{
    shutdown(icc_server_socket_id, SHUT_RDWR);
    pthread_join(icc_server_host_thread, NULL);
}