#include "../internal.h"
#include "../mcumgr.h"
#include "COD_Compiler_Defines.h"
#include "lion/log/log.h"
#include "lion/utils/byteorder.h"
#include "zcbor_decode.h"
#include <argp.h>
#include <assert.h>
#include <fcntl.h>
#include <openssl/evp.h>
#include <openssl/sha.h>
#include <stdlib.h>

LOG_MODULE_REGISTER(image, LOG_LEVEL_NONE);

static const char *APP_IMG_Upload_szFirmware;

static struct argp_option args_options[] = {{0}};

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    char *endptr;
    switch (key)
    {
        case ARGP_KEY_END:
            if (state->arg_num < 1)
            {
                argp_error(state, "Too few arguments!");
            }
            break;
        case ARGP_KEY_ARG:
            switch (state->arg_num)
            {
                case 0:
                    APP_IMG_Upload_szFirmware = arg;
                    break;
                default:
                    argp_usage(state);
            }
            break;
        default:
            return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static struct argp args_argp = {args_options, parse_opt, "FILE", ""};

int APP_IMG_Upload_Main(int argc, char *argv[])
{
    argp_parse(&args_argp, argc, argv, 0, 0, 0);

    FILE *fdFirmware = fopen(APP_IMG_Upload_szFirmware, "r");
    if (fdFirmware == NULL)
    {
        printf("Unable to open \"%s\"\n", APP_IMG_Upload_szFirmware);
        return EIO;
    }

    printf("Uploading \"%s\"\n", APP_IMG_Upload_szFirmware);

    size_t  sizTotal = 0;
    uint8_t arrSHA256[32];

    StackScop EVP_MD_CTX *ptr_EVPContext = EVP_MD_CTX_new();
    EVP_DigestInit_ex(ptr_EVPContext, EVP_sha256(), NULL);
    StackScop uint8_t arrReadBuffer[4096];
    StackScop size_t  sizRead;
    while ((sizRead = fread(arrReadBuffer, 1, sizeof(arrReadBuffer), fdFirmware)))
    {
        EVP_DigestUpdate(ptr_EVPContext, arrReadBuffer, sizRead);
        sizTotal += sizRead;
    }
    unsigned int int_DigestLength = sizeof(arrSHA256);
    EVP_DigestFinal_ex(ptr_EVPContext, arrSHA256, &int_DigestLength);
    fseek(fdFirmware, 0, SEEK_SET);

    printf("Length: %zu bytes\n", sizTotal);
    char szSHA256[(SHA256_DIGEST_LENGTH * 2) + 1];
    for (uint8_t u8Index = 0; u8Index < sizeof(arrSHA256); u8Index++)
    {
        assert(sprintf(&szSHA256[u8Index * 2], "%02X", arrSHA256[u8Index]) == 2);
    }
    szSHA256[sizeof(szSHA256) - 1U] = '\0';
    printf("SHA256: %s\n", szSHA256);

    uint8_t       arrBuffer[2044U - sizeof(MCUMGR_Header_t)];
    zcbor_state_t arrZCBORState[4] = {0};

    size_t sizOffset  = 0;
    bool   bolMatches = false;
    while (sizOffset < sizTotal)
    {
        zcbor_new_encode_state(arrZCBORState, 4, arrBuffer, sizeof(arrBuffer), 1);
        assert(zcbor_map_start_encode(arrZCBORState, 6));

        assert(zcbor_tstr_put_lit(arrZCBORState, "off"));
        assert(zcbor_uint32_put(arrZCBORState, sizOffset));
        if (sizOffset == 0)
        {
            assert(zcbor_tstr_put_lit(arrZCBORState, "len"));
            assert(zcbor_uint32_put(arrZCBORState, sizTotal));
            assert(zcbor_tstr_put_lit(arrZCBORState, "sha"));
            assert(zcbor_bstr_put_arr(arrZCBORState, arrSHA256));
            /* Setting upgrade to true forces version comparison. */
            assert(zcbor_tstr_put_lit(arrZCBORState, "upgrade"));
            assert(zcbor_bool_put(arrZCBORState, true));
        }
        assert(zcbor_tstr_put_lit(arrZCBORState, "data"));

        /* Add 1 for map end, get maximum length of bstr header. */
        /* cppcheck-suppress misra-c2012-18.4 - Don't have a choice */
        size_t sizUsedLength      = (arrZCBORState[0].payload - arrBuffer) + 1U;
        size_t sizAvailableLength = sizeof(arrBuffer) - sizUsedLength;
        size_t sizToReadLength    = sizAvailableLength;
        while (sizToReadLength + zcbor_header_len(sizToReadLength) > sizAvailableLength)
        {
            sizToReadLength--;
            assert(sizToReadLength > 0U);
        }
        size_t sizReadLength = fread(arrReadBuffer, 1, sizToReadLength, fdFirmware);
        assert(zcbor_bstr_encode_ptr(arrZCBORState, arrReadBuffer, sizReadLength));
        assert(zcbor_map_end_encode(arrZCBORState, 6));

        size_t sizPayloadLength = (arrZCBORState[0].payload - arrBuffer);
        printf("Writing %zu bytes at 0x%08zX (%zu bytes)\n", sizReadLength, sizOffset, sizOffset);
        fflush(stdout);
        uint16_t u16ResponseLength = sizeof(arrBuffer);
        int      intResult = app_mcumgr_Command(MCUMGR_OP_u8WRITE, MCUMGR_IMAGE_u16GROUP, MCUMGR_IMAGE_u8ID_UPLOAD,
                                                arrBuffer, sizPayloadLength, arrBuffer, &u16ResponseLength, RTOS_SECONDS(5));

        if (intResult < 0)
        {
            printf("MCUMGR Execution Error %d \"%s\"", intResult, strerror(-intResult));
            return EIO;
        }

        sizOffset += sizReadLength;

        zcbor_new_decode_state(arrZCBORState, 4, arrBuffer, u16ResponseLength, 1);
        assert(zcbor_map_start_decode(arrZCBORState));
        struct zcbor_string strKey;
        uint32_t            u32Offset;
        uint32_t            u32_RC;
        while (zcbor_tstr_decode(arrZCBORState, &strKey))
        {
            if (strncmp("off", strKey.value, strKey.len) == 0)
            {
                zcbor_uint32_decode(arrZCBORState, &u32Offset);
                if (sizOffset != u32Offset)
                {
                    printf("Seeking to reported offset (%u bytes)\n", u32Offset);
                    sizOffset = u32Offset;
                    fseek(fdFirmware, sizOffset, SEEK_SET);
                }
            }
            else if (strncmp("matches", strKey.value, strKey.len) == 0)
            {
                zcbor_bool_decode(arrZCBORState, &bolMatches);
                printf("Firmware reported SHA256 match: %c\n", bolMatches ? 'Y' : 'N');
            }
            else if (app_mcumgr_ParseRet(arrZCBORState, &strKey, &u32_RC))
            {
                if (u32_RC != 0U)
                {
                    printf("Firmware reported error %u\n", u32_RC);
                    return u32_RC;
                }
            }
            else
            {
                char *szKey = malloc(strKey.len + 1);
                strncpy(szKey, strKey.value, strKey.len);
                LOG_WRN("Unknown key: \"%s\"", szKey);
                free(szKey);
                zcbor_any_skip(arrZCBORState, NULL);
            }
        }
        assert(zcbor_map_end_decode(arrZCBORState));
    }

    fclose(fdFirmware);

    if (sizOffset == sizTotal && bolMatches)
    {
        printf("Success!\n");
        return 0;
    }
    else
    {
        return EIO;
    }
}