#include "lion/icc/ICC.h"
#include "lion/log/log.h"
#include "mcumgr.h"
#include "zcbor_decode.h"
#include "zcbor_encode.h"
#include <argp.h>
#include <errno.h>
#include <lion/utils/byteorder.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <termios.h>

#include "img_mgmt/img_mgmt.h"
#include "internal.h"
#include "os_mgmt/os_mgmt.h"

LOG_MODULE_REGISTER(app, LOG_LEVEL_NONE);

static ICC_Event_Handler_t app_strICCResponseHandler;

static sem_t     app_strCommandLock; // Mutex?
static uint8_t   app_u8CurrentOp;
static uint16_t  app_u16CurrentGroup;
static uint8_t   app_u8CurrentId;
static sem_t     app_strResponse;
static void     *app_ptrResponseData;
static uint16_t *app_ptrResponseLength;
static int       app_intRetval;

static uint32_t app_u32ICCFunction = 0;

int app_mcumgr_Command(uint8_t u8Op, uint16_t u16Group, uint8_t u8Id, const void *ptrData, uint16_t u16DataLength,
                       uint8_t *ptrResponse, uint16_t *u16ResponseLength, RTOS_Timeout_t strTimeout)
{
    assert(((ptrResponse != NULL) && (u16ResponseLength != NULL)) ||
           ((ptrResponse == NULL) &&
            u16ResponseLength == NULL)); /* ptrResponse and ptrResponseLength must both be Non-NULL, or NULL */
    assert((ptrData != NULL) || (u16DataLength == 0)); /* u16DataLength MUST be 0 if ptrData is NULL */

    MCUMGR_Header_t header = {.nh_op    = u8Op,
                              .nh_flags = 0,
                              .nh_len   = (ptrData == NULL) ? 0 : sys_cpu_to_be16(u16DataLength),
                              .nh_group = sys_cpu_to_be16(u16Group),
                              .nh_seq   = 0,
                              .nh_id    = u8Id};

    uint32_t u32ResponseLength = u16ResponseLength == NULL ? 4096 : *u16ResponseLength;
    uint32_t u32TotalLength    = sizeof(MCUMGR_Header_t) + u16DataLength;
    uint8_t *arrRequestBuffer  = malloc(u32TotalLength);
    uint8_t *arrResponseBuffer = malloc(u32ResponseLength);
    memcpy(arrRequestBuffer, &header, sizeof(MCUMGR_Header_t));
    if (ptrData != NULL)
    {
        memcpy(&arrRequestBuffer[sizeof(MCUMGR_Header_t)], ptrData, u16DataLength);
    }
    int intReturnvalue = ICC_Function_Call(&APP_strICC, app_u32ICCFunction, arrRequestBuffer, u32TotalLength,
                                           arrResponseBuffer, &u32ResponseLength, strTimeout);
    free(arrRequestBuffer);
    if (u32ResponseLength >= sizeof(MCUMGR_Header_t))
    {
        if (ptrResponse != NULL)
        {
            memcpy(ptrResponse, &arrResponseBuffer[sizeof(MCUMGR_Header_t)],
                   u32ResponseLength - sizeof(MCUMGR_Header_t));
        }
        if (u16ResponseLength != NULL)
        {
            *u16ResponseLength = u32ResponseLength - sizeof(MCUMGR_Header_t);
        }
    }
    else if (u16ResponseLength != NULL)
    {
        *u16ResponseLength = 0;
    }
    free(arrResponseBuffer);

    return intReturnvalue;
}

bool app_mcumgr_ParseRet(zcbor_state_t *ptr_CBOR, struct zcbor_string *ptr_Key, uint32_t *ptr_RC)
{
    *ptr_RC = 0;
    if (strncmp("err", ptr_Key->value, ptr_Key->len) == 0)
    {
        zcbor_map_start_decode(ptr_CBOR);
        struct zcbor_string str_InnerKey;
        while (zcbor_tstr_decode(ptr_CBOR, &str_InnerKey))
        {
            if (strncmp("group", str_InnerKey.value, str_InnerKey.len) == 0)
            {
                zcbor_any_skip(ptr_CBOR, NULL);
            }
            else if (strncmp("rc", str_InnerKey.value, str_InnerKey.len) == 0)
            {
                zcbor_uint32_decode(ptr_CBOR, ptr_RC);
            }
            else
            {
                char *sz_Key = malloc(str_InnerKey.len + 1);
                strncpy(sz_Key, str_InnerKey.value, str_InnerKey.len);
                LOG_WRN("Unknown key: \"%s\"", sz_Key);
                free(sz_Key);
                zcbor_any_skip(ptr_CBOR, NULL);
            }
        }
        zcbor_map_end_decode(ptr_CBOR);
        return true;
    }
    else if (strncmp("rc", ptr_Key->value, ptr_Key->len) == 0)
    {
        zcbor_uint32_decode(ptr_CBOR, ptr_RC);
        return true;
    }
    else
    {
        return false;
    }
}

static struct argp_option args_options[] = {
    {.name  = "device",
     .key   = 'd',
     .arg   = "device-code",
     .flags = 0,
     .doc   = "The ICC device to access. Either a string identifier, or an explicit ICC function in hex or decimal.",
     .group = 0},
    {0}
};

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    char *endptr;
    switch (key)
    {
        case 'd':
        {
            int intBase = 10;
            if (strncmp(arg, "0x", 2) == 0 || strncmp(arg, "0X", 2) == 0)
            {
                intBase = 16;
            }
            char *ptrEndPtr;
            app_u32ICCFunction = (uint32_t)strtoul(arg, &ptrEndPtr, intBase);
            if ((ptrEndPtr - arg) != strlen(arg))
            {
                if (strcmp(arg, "dcu") == 0)
                {
                    app_u32ICCFunction = 0x02FF0000;
                }
                else if (strcmp(arg, "pmu") == 0)
                {
                    app_u32ICCFunction = 0x03FF0000;
                }
                else
                {
                    argp_error(state, "\"%s\" is not a valid integer or identifier", arg);
                }
            }
            break;
        }
        break;
        case ARGP_KEY_END:
            break;
        case ARGP_KEY_ARG:
            return ARGP_ERR_UNKNOWN;
        default:
            return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static struct argp args_argp = {args_options, parse_opt, "", ""};

int main(int argc, char *argv[])
{
    int retval = -1;

    sem_init(&app_strCommandLock, 0, 1);
    sem_init(&app_strResponse, 0, 1);

    int argindex = 23;
    argp_parse(&args_argp, argc, argv, ARGP_IN_ORDER, &argindex, 0);
    argc = argc - argindex;
    argv = &argv[argindex];

    LOG_INF("Using ICC Function ID: 0x%X", app_u32ICCFunction);

    APP_ICC_Init();

    if (APP_ICC_Start() != 0)
    {
        printf("Unable to connect to ICC server\n");
        retval = EIO;
    }
    else if (argc >= 1)
    {
        if (strcmp(argv[0], "image") == 0)
        {
            if (argc >= 2)
            {
                if (strcmp(argv[1], "upload") == 0)
                {
                    retval = APP_IMG_Upload_Main(argc - 1, &argv[1]);
                }
                else if (strcmp(argv[1], "list") == 0)
                {
                    retval = APP_IMG_List_Main(argc - 1, &argv[1]);
                }
                else if (strcmp(argv[1], "erase") == 0)
                {
                    retval = APP_IMG_Erase_Main(argc - 1, &argv[1]);
                }
                else if (strcmp(argv[1], "test") == 0)
                {
                    retval = APP_IMG_Test_Main(argc - 1, &argv[1]);
                }
                else if (strcmp(argv[1], "confirm") == 0)
                {
                    retval = APP_IMG_Confirm_Main(argc - 1, &argv[1]);
                }
                else
                {
                    printf("Unknown Command %s\n", argv[1]);
                }
            }
            else
            {
                printf("Unknown Command\n");
            }
        }
        else if (strcmp(argv[0], "os") == 0)
        {

            if (strcmp(argv[1], "reset") == 0)
            {
                retval = APP_OS_Reset_Main(argc - 1, &argv[1]);
            }
            else
            {
                printf("Unknown Command %s\n", argv[1]);
            }
        }
        else
        {
            printf("Unknown Command %s\n", argv[0]);
        }
    }
    else
    {
        printf("Unknown Command\n");
    }

    APP_ICC_Stop();
    fflush(stdout);
    return retval;
}