#ifndef LION_ICC_ICC_H__
#define LION_ICC_ICC_H__

/* ======================================================================= */
/* =========================== I N C L U D E S =========================== */
/* ======================================================================= */

#include "COD_Compiler_Defines.h"
#include "lion/icc/ICC_Config.h"
#include "lion/icc/ICC_Event.h"
#include "lion/icc/ICC_Function.h"
#include "lion/icc/ICC_Transport.h"
#include "lion/rtos/RTOS_CondVar.h"
#include "lion/rtos/RTOS_Mutex.h"
#include "lion/rtos/RTOS_Sem.h"
#include "lion/utils/dlist.h"
#include <assert.h>
#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* ======================================================================= */
    /* ============================= D E F I N E S =========================== */
    /* ======================================================================= */

#define icc_u32BUFFER_COUNT (CONFIG_ICC_u32RECV_BUFFER_COUNT + CONFIG_ICC_u32XMIT_BUFFER_COUNT)
    static_assert(icc_u32BUFFER_COUNT <= 32U, "Too many buffers declared");

#define icc_u8PACKET_TYPE_RETURNV2 ((uint8_t)0x0)
#define icc_u8PACKET_TYPE_EVENT    ((uint8_t)0x1)
#define icc_u8PACKET_TYPE_FUNCTION ((uint8_t)0x2)
#define icc_u8PACKET_TYPE_RETURN   ((uint8_t)0x3) /** @deprecated To be removed when no longer used. */
#define icc_u8PACKET_TYPE_SYSTEM   ((uint8_t)0x3) /** @warning DO NOT USE until icc_u8PACKET_TYPE_RETURN is removed. */

    /* ======================================================================= */
    /* ============================ T Y P E D E F S ========================== */
    /* ======================================================================= */

    typedef struct
    {
        uint8_t u3Type        : 3;
        uint8_t u2Transaction : 2;
        uint8_t u2_Reserved   : 2;
        uint8_t flgEncrypted  : 1;
    } icc_PacketHeader_t;

    typedef struct __attribute__((packed))
    {
        icc_PacketHeader_t strHeader;
        uint32_t           u32FunctionId;
    } icc_FunctionHeader_t;

    /** @deprecated To be removed when no longer used. */
    typedef struct __attribute__((packed))
    {
        icc_PacketHeader_t strHeader;
        uint32_t           u32FunctionId;
        int32_t            i32ReturnValue;
    } icc_ReturnHeader_t;

    typedef struct __attribute__((packed))
    {
        icc_PacketHeader_t strHeader;
        uint32_t           u32FunctionId;
        int32_t            i32ReturnValue;
        uint16_t           u16_RequestCRC;
    } icc_ReturnHeaderV2_t;

    typedef struct __attribute__((packed))
    {
        icc_PacketHeader_t str_Header;
        uint32_t           u32_EventId;
    } icc_EventHeader_t;

#define icc_MIN(a, b) ((a) < (b) ? (a) : (b))
#define icc_MAX(a, b) ((a) > (b) ? (a) : (b))
#define icc_u32PACKET_OVERHEAD        \
    ((uint32_t)(icc_MAX(              \
        sizeof(icc_FunctionHeader_t), \
        icc_MAX(sizeof(icc_ReturnHeaderV2_t), icc_MAX(sizeof(icc_ReturnHeader_t), sizeof(icc_EventHeader_t))))))
#define icc_u32RECV_BUFFER_SIZE (CONFIG_ICC_u32RECV_BUFFER_SIZE + icc_u32PACKET_OVERHEAD)
#define icc_u32XMIT_BUFFER_SIZE (CONFIG_ICC_u32XMIT_BUFFER_SIZE + icc_u32PACKET_OVERHEAD)

    struct ICC_Packet_t
    {
        uint8_t          u8Index;
        sys_dnode_t      node;
        ICC_Transport_t *ptrSource; /* Specifies which transport the packet came from, or NULL if local. */
        ICC_Transport_t *ptrTarget; /* Specifies which transport to send the packet to, or NULL for broadcast. */
        uint16_t         u16Length;
        uint8_t         *arrBuffer;
    };
    typedef struct ICC_Packet_t ICC_Packet_t;

    struct ICC_Context_t
    {
        uint8_t      arrRecvBuffers[CONFIG_ICC_u32RECV_BUFFER_COUNT][icc_u32RECV_BUFFER_SIZE];
        uint8_t      arrXmitBuffers[CONFIG_ICC_u32XMIT_BUFFER_COUNT][icc_u32XMIT_BUFFER_SIZE];
        ICC_Packet_t arrPackets[icc_u32BUFFER_COUNT];
        uint32_t     u32PacketUsage;
        RTOS_Sem_t   strRecvSemaphore;
        RTOS_Sem_t   strXmitSemaphore;

        RTOS_Mutex_t strHandlerLock;
        RTOS_Mutex_t mutex;
        sys_dlist_t  events;
        sys_dlist_t  functions;

        RTOS_Mutex_t strTransportLock;
        sys_dlist_t  transports;

        sys_dlist_t strFunctionCalls;

        RTOS_Mutex_t   mtx_QueueLock;
        RTOS_CondVar_t cnd_QueueVar;
        sys_dlist_t    packetQueue;
        uint8_t        u8_Transaction;
    };
    typedef struct ICC_Context_t ICC_Context_t;

    /* ======================================================================= */
    /* ========================== P R O T O T Y P E S ======================== */
    /* ======================================================================= */

    void ICC_Init(ICC_Context_t *context);

    /**
     * @brief Processes all queued packets, and then all of the transports.
     * @param ptr_Context The ICC Context
     * @warning Use @ref ICC_ProcessIO and @ref ICC_ProcessPacket in multiple threads instead if long running function
     * calls are expected.
     * @note This should be called every 10ms
     */
    GlobScop void ICC_Process(ICC_Context_t *ptr_Context);

    /**
     * @brief Process all of the transports.
     * @param ptr_Context The ICC Context
     * @note This should be called every 10ms
     */
    GlobScop void ICC_ProcessIO(ICC_Context_t *ptr_Context);

    /**
     * @brief Process one packet from the packet queue.
     * @param ptr_Context The ICC Context
     * @param tim_Timeout Time to wait for a single packet to process.
     * @return True is a packet was processed, or false if not.
     *
     * Any number of threads may be spooled up to call this function repeatedly. The simplest of which looks like:
     * @code while(true) { ICC_ProcessPacket(&ICCContext, RTOS_FOREVER); } @endcode
     */
    GlobScop bool ICC_ProcessPacket(ICC_Context_t *ptr_Context, RTOS_Timeout_t tim_Timeout);

#ifdef __cplusplus
}
#endif

#endif /* LION_ICC_ICC_H__ */