#ifndef LION_ICC_EVENTS_H__
#define LION_ICC_EVENTS_H__

/* ======================================================================= */
/* =========================== I N C L U D E S =========================== */
/* ======================================================================= */

#include "lion/utils/dlist.h"
#include "lion/rtos/RTOS_Timeout.h"
#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* ======================================================================= */
    /* ============================= D E F I N E S =========================== */
    /* ======================================================================= */

    /* ======================================================================= */
    /* ============================ T Y P E D E F S ========================== */
    /* ======================================================================= */

    typedef struct ICC_Context_t ICC_Context_t;

    typedef void (*ICC_Event_Callback_t)(int32_t eventId, const void *eventData, size_t eventDataLength,
                                         void *userData);

    typedef struct
    {
        sys_dlist_t          node;
        uint32_t             u32_EventID;
        uint32_t             u32_EventMask;
        ICC_Event_Callback_t fn_Callback;
        void                *ptr_UserData;
    } ICC_Event_Handler_t;

    /**
     * @brief Initialize an ICC_Event_Handler_t
     * @param ptr_Handler A reference an ICC_Event_Handler_t structure allocated by the user
     * @param u32_EventID The eventId this handler will be called for
     *
     * @param fn_Callback The function that will be called when the event occurs
     * @param ptr_UserData An arbitrary pointer that will be passed to the callback function
     */
    void ICC_Event_InitHandler(ICC_Event_Handler_t *ptr_Handler, uint32_t u32_EventID, uint32_t u32_EventMask,
                               ICC_Event_Callback_t fn_Callback, void *ptr_UserData);

    /**
     * @brief Registers a handler with the event subsystem
     * @warning handler MUST remain valid until ICC_Event_UnregisterHandler is called
     * @param ptr_ICC Pointer to target ICC context
     * @param ptr_Handler An initialized ICC_Event_Handler_t structure
     * @returns 0 on Success
     * @returns -EALREADY if handler is already registered
     */
    int ICC_Event_RegisterHandler(ICC_Context_t *ptr_ICC, ICC_Event_Handler_t *ptr_Handler);

    /**
     * @brief Unregisters a handler from the event subsystem
     * @param ptr_ICC Pointer to target ICC context
     * @param ptr_Handler An initialized ICC_Event_Handler_t structure
     * @returns 0 on Success
     * @returns -EALREADY if handler is already unregistered
     * @warning Currently Unsafe to Call: Mutex doesn't properly lock the list during packet processing.
     */
    int ICC_Event_UnregisterHandler(ICC_Context_t *ptr_ICC, ICC_Event_Handler_t *ptr_Handler);

    /**
     * @brief Dispatches an event to all event handlers.
     * @param ptr_ICC Pointer to target ICC context
     * @param eventId The event ID
     * @param ptr_EventData A pointer to the data to be passed to event handlers
     * @param ptr_DataLength The length in bytes of eventData
     * @param tim_Timeout Timeout to wait until event is allocated and dispatched
     * @returns 0 on Success
     * @returns -EINVAL If an illegal event ID is used
     * @returns -EAGAIN If timeout expires waiting for queue to unblock, or queue is full when called from ISR
     * @returns -ENOMEM If eventDataLength is too large
     */
    int ICC_Event_Dispatch(ICC_Context_t *ptr_ICC, uint32_t u32_EventID, const void *ptr_EventData,
                           size_t ptr_DataLength, RTOS_Timeout_t tim_Timeout);

#ifdef __cplusplus
}
#endif

#endif /* LION_ICC_EVENTS_H__ */