#ifndef LION_ICC_FUNCTIONS_H__
#define LION_ICC_FUNCTIONS_H__

/* ======================================================================= */
/* =========================== I N C L U D E S =========================== */
/* ======================================================================= */

#include "lion/rtos/RTOS_Timeout.h"
#include "lion/utils/dlist.h"
#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

    /* ======================================================================= */
    /* ============================= D E F I N E S =========================== */
    /* ======================================================================= */

#define ICC_Function_ANY_ID -1

    /* ======================================================================= */
    /* ============================ T Y P E D E F S ========================== */
    /* ======================================================================= */

    typedef struct ICC_Context_t ICC_Context_t;

    typedef int (*ICC_Function_HandlerCallback_t)(uint32_t u32Id, const void *ptrData, uint32_t u32DataLength,
                                                  void *ptrResponse, uint32_t *u32ResponseLength, void *ptrUserData);

    typedef void (*ICC_Function_Callback_t)(uint32_t u32FunctionId, int intResult, const void *ptrResponse,
                                            uint32_t u32ResponseLength, void *ptrUserData);
    typedef struct
    {
        sys_dlist_t                    node;
        uint32_t                       u32FunctionId;
        uint32_t                       u32FunctionIdMask;
        ICC_Function_HandlerCallback_t fnCallback;
        void                          *ptrUserData;
    } ICC_Function_Handler_t;

    struct ICC_FunctionCall_t
    {
        ICC_Context_t          *ptr_ICCContext;
        sys_dnode_t             strNode;
        uint32_t                u32FunctionId;
        uint16_t                u16_RequestCRC;
        ICC_Function_Callback_t fnCallback;
        void                   *ptrUserData;
    };
    typedef struct ICC_FunctionCall_t ICC_FunctionCall_t;

    /**
     * @brief Initialize an ICC_Function_Handler_t
     * @param handler A reference an ICC_Function_Handler_t structure allocated by the user
     * @param functionId The functionId this handler will be called for
     * @param functionIdMask A mask that is first applied to the functionId before checking for a match
     * @param callback The function that will be called when the event occurs
     * @param userData An arbitrary pointer that will be passed to the callback function
     */
    void ICC_Function_InitHandler(ICC_Function_Handler_t *handler, uint32_t functionId, uint32_t functionIdMask,
                                  ICC_Function_HandlerCallback_t callback, void *userData);

    /**
     * @brief Registers a handler with the function subsystem
     * @warning handler must remain valid until ICC_Function_UnregisterHandler is called
     * @param handler An initialized ICC_Function_Handler_t structure
     * @returns 0 on Success
     * @returns -EALREADY if handler is already registered, or the ID + mask combo conflicts with another handler
     */
    int ICC_Function_RegisterHandler(ICC_Context_t *context, ICC_Function_Handler_t *handler);

    /**
     * @brief Unregisters a handler from the function subsystem
     * @param handler An initialized ICC_Function_Handler_t structure
     * @returns 0 on Success
     * @returns -EALREADY if handler is already unregistered
     * @warning Currently Unsafe to Call: Mutex doesn't properly lock the list during packet processing.
     */
    int ICC_Function_UnregisterHandler(ICC_Context_t *context, ICC_Function_Handler_t *handler);

    /**
     * @brief Dispatches a function to the first encountered function handler. (Local First)
     * @param functionId The function ID
     * @param data A pointer to the data to be passed to the function handler
     * @param dataLength The length in bytes of data
     * @param timeoutMS Milliseconds to block on a full function queue
     * @returns 0 on Success
     * @returns -EINVAL If an illegal event ID is used
     * @returns -EAGAIN If timeout expires waiting for queue to unblock, or queue is full when called from ISR
     * @returns -ENOMEM If eventDataLength is too large
     */
    int ICC_Function_Call(ICC_Context_t *ptrContext, uint32_t u32FunctionId, const void *ptrData,
                          uint32_t u32DataLength, void *ptrResponse, uint32_t *u32ResponseLength,
                          RTOS_Timeout_t strTimeout);

    int ICC_Function_CallAsync(ICC_Context_t *ptrContext, ICC_FunctionCall_t *ptrFunctionCall, uint32_t u32FunctionId,
                               const void *ptrData, uint32_t u32DataLength, ICC_Function_Callback_t fnCallback,
                               void *ptrUserData);

    void ICC_Function_InitCall(ICC_FunctionCall_t *ptrFunctionCall);

    /**
     * @brief Destroys the provided function call context.
     * @param context
     * @param ptrFunctionCall
     * @return
     */
    void ICC_Function_DestroyCall(ICC_FunctionCall_t *ptrFunctionCall);

#ifdef __cplusplus
}
#endif

#endif /* LION_ICC_FUNCTIONS_H__ */