/* ======================================================================= */
/* =========================== I N C L U D E S =========================== */
/* ======================================================================= */

#include "lion/icc/transport/ICC_Stream.h"
#include "ICC_Internal.h"
#include "lion/icc/ICC.h"
#include "lion/log/log.h"
#include "lion/utils/byteorder.h"
#include "lion/utils/crc.h"
#include <stdalign.h>
#include <string.h>
#ifndef __ZEPHYR__
    #include <stdlib.h>
#endif

LOG_MODULE_REGISTER(icc_stream, CONFIG_ICC_LOG_LEVEL);

/**********************************************************************************************************************
 * @cond                                              FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

FileScop const uint8_t icc_stream_Delimiter[] = {CONFIG_ICC_STREAM_DELIMITER};

FileScop bool icc_stream_HasTXPacket(ICC_Stream_Context_t *context) {
    if (context->tx_packet == NULL) {
        context->tx_packet = ICC_Transport_PopTXPacket(&context->transport);
    }
    return context->tx_packet != NULL;
}

FileScop void icc_stream_ProcessTX(ICC_Stream_Context_t *context)
{
    uint16_t temp_length;
    uint16_t temp_crc;
    bool     bolInsufficientSpace = false;
    uint32_t u32PushedBytes;
    StackScop uint32_t u32_Iterations = 0;

    while (icc_stream_HasTXPacket(context) && !bolInsufficientSpace)
    {
        u32_Iterations++;
        switch (context->tx_state)
        {
            case ICC_STREAM_TX_PUT_SYNC:
                if (context->fnWrite(icc_stream_Delimiter, sizeof(icc_stream_Delimiter), true, context->ptrUserData) ==
                    sizeof(icc_stream_Delimiter))
                {
                    context->tx_state         = ICC_STREAM_TX_PUT_LENGTH;
                    context->tx_calculatedCRC = 0x0000;
                }
                else
                {
                    bolInsufficientSpace = true;
                }
                break;
            case ICC_STREAM_TX_PUT_LENGTH:
                temp_length = sys_cpu_to_le16(context->tx_packet->u16Length);
                if (context->fnWrite((const uint8_t *)&temp_length, sizeof(uint16_t), true, context->ptrUserData) ==
                    sizeof(uint16_t))
                {
                    context->tx_frameCounter = 0;
                    context->tx_state        = ICC_STREAM_TX_PUT_PACKET;
                    context->tx_calculatedCRC =
                        crc16_ccitt(context->tx_calculatedCRC, (uint8_t *)&temp_length, sizeof(uint16_t));
                }
                else
                {
                    bolInsufficientSpace = true;
                }
                break;
            case ICC_STREAM_TX_PUT_PACKET:
                u32PushedBytes = context->fnWrite(&context->tx_packet->arrBuffer[context->tx_frameCounter],
                                                  context->tx_packet->u16Length - context->tx_frameCounter, false,
                                                  context->ptrUserData);
                if (u32PushedBytes > 0)
                {
                    context->tx_frameCounter += u32PushedBytes;
                }
                else
                {
                    bolInsufficientSpace = true;
                }
                if (context->tx_frameCounter >= context->tx_packet->u16Length)
                {
                    context->tx_state = ICC_STREAM_TX_PUT_CRC;
                }
                break;
            case ICC_STREAM_TX_PUT_CRC:
                temp_crc = sys_cpu_to_le16(context->tx_calculatedCRC);
                if (context->fnWrite((const uint8_t *)&temp_crc, sizeof(uint16_t), true, context->ptrUserData) ==
                    sizeof(uint16_t))
                {
                    ICC_Transport_ReleaseTXPacket(&context->transport, context->tx_packet);
                    context->tx_packet = NULL;
                    context->tx_state  = ICC_STREAM_TX_PUT_SYNC;
                }
                else
                {
                    bolInsufficientSpace = true;
                }
                break;
            default:
                context->tx_frameCounter = 0;
                context->tx_state        = ICC_STREAM_TX_PUT_SYNC;
                break;
        }
    }

    if ((u32_Iterations <= 1U) && (context->tx_packet != NULL)) {
        /* If the loop was only run once or not at all, nothing was transmitted */
        if (++context->u16_TX_Cycles >= 10U)
        {
            LOG_WRN("%p: Packet %p Timedout for TX!", context, context->tx_packet);
            if (context->tx_packet != NULL)
            {
                ICC_Transport_ReleaseTXPacket(&context->transport, context->tx_packet);
                context->tx_packet = NULL;
            }
            context->tx_state = ICC_STREAM_TX_PUT_SYNC;
            context->u16_TX_Cycles = 0;
        }
    } else {
        context->u16_TX_Cycles = 0;
    }
}

FileScop void icc_stream_ProcessRX(ICC_Stream_Context_t *context)
{
    uint8_t  temp_data;
    uint16_t temp_length;
    uint16_t temp_crc;
    uint32_t u32PulledBytes;
    bool     bolEmpty      = false;
    uint32_t u32Iterations = 0;

    while (!bolEmpty)
    {
        u32Iterations++;
        switch (context->rx_state)
        {
            case ICC_STREAM_RX_WAIT_FOR_SYNC:
                if (context->rx_packet != NULL)
                {
                    icc_ReleasePacket(context->transport.context, context->rx_packet);
                    context->rx_packet = NULL;
                }
                if (context->fnRead(&temp_data, 1, true, context->ptrUserData) == 1)
                {
                    if (icc_stream_Delimiter[context->rx_frameCounter] == temp_data)
                    {
                        context->rx_frameCounter++;
                        if (context->rx_frameCounter >= sizeof(icc_stream_Delimiter))
                        {
                            // Go to next state to get frame length
                            context->rx_state         = ICC_STREAM_RX_GET_LENGTH;
                            // TODO: Reset Timer Counter
                            context->rx_calculatedCRC = 0x0000;
                        }
                    }
                    else
                    {
                        // Discard data
                        context->rx_frameCounter = 0;
                    }
                }
                else
                {
                    bolEmpty = true;
                }
                break;
            case ICC_STREAM_RX_GET_LENGTH:
                if (context->fnRead((uint8_t *)&temp_length, sizeof(uint16_t), true, context->ptrUserData) ==
                    sizeof(uint16_t))
                {
                    context->rx_frameLength = sys_le16_to_cpu(temp_length);
                    context->rx_frameCounter += sizeof(uint16_t);
                    context->rx_calculatedCRC =
                        crc16_ccitt(context->rx_calculatedCRC, (uint8_t *)&temp_length, sizeof(uint16_t));
                    context->rx_state = ICC_STREAM_RX_ALLOCATE_PACKET;
                }
                else
                {
                    bolEmpty = true;
                }
                break;
            case ICC_STREAM_RX_ALLOCATE_PACKET:
                context->rx_packet = icc_AllocateRXPacket(context->transport.context, context->rx_frameLength);
                if (context->rx_packet == NULL)
                {
                    bolEmpty = true;
                }
                else
                {
                    context->rx_frameCounter = 0;
                    context->rx_state        = ICC_STREAM_RX_GET_PACKET;
                }
                break;

            case ICC_STREAM_RX_GET_PACKET:
                u32PulledBytes =
                    context->fnRead(&context->rx_packet->arrBuffer[context->rx_frameCounter],
                                    context->rx_frameLength - context->rx_frameCounter, false, context->ptrUserData);
                if (u32PulledBytes > 0)
                {
                    context->rx_frameCounter += u32PulledBytes;
                }
                else
                {
                    bolEmpty = true;
                }
                if (context->rx_frameCounter >= context->rx_frameLength)
                {
                    context->rx_state = ICC_STREAM_RX_GET_CRC;
                }
                break;
            case ICC_STREAM_RX_GET_CRC:
                if (context->fnRead((uint8_t *)&temp_crc, sizeof(uint16_t), true, context->ptrUserData) ==
                    sizeof(uint16_t))
                {
                    if (context->rx_calculatedCRC == temp_crc)
                    {
                        context->rx_packet->ptrSource = &context->transport;
                        context->rx_packet->u16Length = context->rx_frameLength;
                        // TODO: Finish frame and call it a day
                        icc_EnqueuePacket(context->transport.context, context->rx_packet);
                    }
                    else
                    {
                        LOG_ERR("%p: Invalid CRC %04X =! %04X", context, temp_crc, context->rx_calculatedCRC);
                        icc_ReleasePacket(context->transport.context, context->rx_packet);
                    }
                    context->rx_packet       = NULL;
                    context->rx_frameCounter = 0;
                    context->rx_state        = ICC_STREAM_RX_WAIT_FOR_SYNC;
                }
                else
                {
                    bolEmpty = true;
                }
                break;
            default:
                context->rx_frameCounter = 0;
                context->rx_state        = ICC_STREAM_RX_WAIT_FOR_SYNC;
                break;
        }
    }

    if ((context->rx_state != ICC_STREAM_RX_WAIT_FOR_SYNC) && (u32Iterations <= 1U))
    {
        /* If the loop was only run once, no useable data was found. If we aren't just waiting for a sync, increment the
         * timeout. */
        context->u16EmptyRXCycles++;
        if (context->u16EmptyRXCycles >= 10U)
        {
            LOG_WRN("%p: RX Timedout!", context);
            if (context->rx_packet != NULL)
            {
                icc_ReleasePacket(context->transport.context, context->rx_packet);
                context->rx_packet = NULL;
            }
            context->rx_frameCounter  = 0;
            context->rx_state         = ICC_STREAM_RX_WAIT_FOR_SYNC;
            context->u16EmptyRXCycles = 0;
        }
    }
    else
    {
        /* Otherwise, the counter is reset. */
        context->u16EmptyRXCycles = 0;
    }
}

FileScop void icc_stream_Process(ICC_Context_t *iccContext, void *userData)
{
    /* userData is guaranteed to be a pointer to an ICC_Stream_Context_t structure */
    /* cppcheck-suppress misra-c2012-11.5 */
    ICC_Stream_Context_t *bsContext = userData;

    icc_stream_ProcessTX(bsContext);
    icc_stream_ProcessRX(bsContext);
}

FileScop void icc_stream_Reset(ICC_Context_t *iccContext, void *userData)
{
    /* userData is guaranteed to be a pointer to an ICC_Stream_Context_t structure */
    /* cppcheck-suppress misra-c2012-11.5 */
    ICC_Stream_Context_t *bsContext = userData;
    if (bsContext->tx_packet != NULL)
    {
        ICC_Transport_ReleaseTXPacket(&bsContext->transport, bsContext->tx_packet);
        bsContext->tx_packet = NULL;
    }
    if (bsContext->rx_packet != NULL)
    {
        icc_ReleasePacket(iccContext, bsContext->rx_packet);
        bsContext->rx_packet = NULL;
    }
}

/**********************************************************************************************************************
 * @cond                                             GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/

GlobScop void ICC_Stream_Init(ICC_Stream_Context_t *context, ICC_Stream_Write_t fnWrite, ICC_Stream_Read_t fnRead,
                              void *ptrUserData)
{
    memset(context, 0, sizeof(ICC_Stream_Context_t));
    ICC_Transport_Init(&context->transport, icc_stream_Process, icc_stream_Reset, context);
    context->rx_state    = ICC_STREAM_RX_WAIT_FOR_SYNC;
    context->tx_state    = ICC_STREAM_TX_PUT_SYNC;
    context->fnWrite     = fnWrite;
    context->fnRead      = fnRead;
    context->ptrUserData = ptrUserData;
}

#ifndef __ZEPHYR__
ICC_Stream_Context_t *ICC_Stream_Allocate() { return malloc(sizeof(ICC_Stream_Context_t)); }
void                  ICC_Stream_Free(ICC_Stream_Context_t *ptr_Context) { free(ptr_Context); }
#endif