#ifndef LION_RTOS_TIMEOUT_H__
#define LION_RTOS_TIMEOUT_H__

/* ======================================================================= */
/* =========================== I N C L U D E S =========================== */
/* ======================================================================= */

#ifdef __cplusplus
extern "C"
{
#endif

    /* ======================================================================= */
    /* ============================ T Y P E D E F S ========================== */
    /* ======================================================================= */

#ifdef __ZEPHYR__

    #include <zephyr/kernel.h>

    typedef k_timeout_t RTOS_Timeout_t;

    #define RTOS_FOREVER K_FOREVER
    #define RTOS_NO_WAIT K_NO_WAIT

    #define RTOS_TIMEOUT_ABS(t) K_TIMEOUT_ABS_MS(t)
    #define RTOS_MSEC(ms)       K_MSEC(ms)
    #define RTOS_SECONDS(s)     K_SECONDS(s)
    #define RTOS_TIMEOUT_EQ(a, b)     K_TIMEOUT_EQ(a, b)

    typedef k_timepoint_t RTOS_Timepoint_t;

    static inline RTOS_Timepoint_t RTOS_TimepointCalc(RTOS_Timeout_t tim_Timeout)
    {
        return sys_timepoint_calc(tim_Timeout);
    }

    static inline RTOS_Timeout_t RTOS_TimepointToTimeout(RTOS_Timepoint_t tim_Timepoint)
    {
        return sys_timepoint_timeout(tim_Timepoint);
    }

    static inline int RTOS_TimepointCmp(RTOS_Timepoint_t a, RTOS_Timepoint_t b) { return sys_timepoint_cmp(a, b); }

#elif __unix__

    #include <pthread.h>
    #include <stdint.h>

    #define rtos_TIMEOUT_MS_PER_SEC 1000
    #define rtos_TIMEOUT_NS_PER_MS  1000000
    #define rtos_TIMEOUT_NS_PER_SEC 1000000000

typedef int64_t RTOS_Timeout_t;

    #define RTOS_FOREVER          ((RTOS_Timeout_t)(-1))
    #define RTOS_NO_WAIT          ((RTOS_Timeout_t)(0))
    #define RTOS_TIMEOUT_EQ(a, b) ((a) == (b))

typedef uint64_t RTOS_Timepoint_t;

static inline RTOS_Timeout_t RTOS_TIMEOUT_ABS(uint32_t u32Milliseconds)
{
    return RTOS_FOREVER - 1 - (int64_t)u32Milliseconds;
}

static inline RTOS_Timeout_t RTOS_MSEC(uint32_t u32Milliseconds) { return (int64_t)u32Milliseconds; }

static inline RTOS_Timeout_t RTOS_SECONDS(uint32_t u32Seconds)
{
    return (int64_t)(u32Seconds * rtos_TIMEOUT_MS_PER_SEC);
}

static inline struct timespec RTOS_TimeoutToTimespec(RTOS_Timeout_t strTimeout)
{
    struct timespec strTime;
    if (strTimeout >= 0)
    {
        clock_gettime(CLOCK_REALTIME, &strTime);
        strTime.tv_sec += strTimeout / rtos_TIMEOUT_MS_PER_SEC;
        strTime.tv_nsec += (strTimeout % rtos_TIMEOUT_MS_PER_SEC) * rtos_TIMEOUT_NS_PER_MS;
        if (strTime.tv_nsec >= rtos_TIMEOUT_NS_PER_SEC)
        {
            strTime.tv_sec++;
            strTime.tv_nsec -= rtos_TIMEOUT_NS_PER_SEC;
        }
    }
    else
    {
        uint64_t u64AbsoluteMS = RTOS_FOREVER - 1 - strTimeout;
        strTime.tv_sec         = u64AbsoluteMS / rtos_TIMEOUT_MS_PER_SEC;
        strTime.tv_nsec        = (u64AbsoluteMS % rtos_TIMEOUT_MS_PER_SEC) * rtos_TIMEOUT_NS_PER_MS;
    }
    return strTime;
}

static inline RTOS_Timepoint_t RTOS_TimepointCalc(RTOS_Timeout_t tim_Timeout)
{
    if (RTOS_TIMEOUT_EQ(tim_Timeout, RTOS_FOREVER))
    {
        return UINT64_MAX;
    }
    else if (RTOS_TIMEOUT_EQ(tim_Timeout, RTOS_NO_WAIT))
    {
        return 0;
    }
    else
    {
        if (tim_Timeout >= 0)
        {
            struct timespec str_Time;
            clock_gettime(CLOCK_REALTIME, &str_Time);
            uint64_t u64_Now =
                (str_Time.tv_sec * rtos_TIMEOUT_MS_PER_SEC) + (str_Time.tv_nsec / rtos_TIMEOUT_NS_PER_MS);
            return u64_Now + tim_Timeout;
        }
        else
        {
            return RTOS_FOREVER - 1 - tim_Timeout;
        }
    }
}

static inline RTOS_Timeout_t RTOS_TimepointToTimeout(RTOS_Timepoint_t tim_Timepoint)
{
    if (tim_Timepoint == UINT64_MAX)
    {
        return RTOS_FOREVER;
    }
    if (tim_Timepoint == 0)
    {
        return RTOS_NO_WAIT;
    }
    struct timespec str_Time;
    clock_gettime(CLOCK_REALTIME, &str_Time);
    uint64_t u64_Now = (str_Time.tv_sec * rtos_TIMEOUT_MS_PER_SEC) + (str_Time.tv_nsec / rtos_TIMEOUT_NS_PER_MS);
    return (tim_Timepoint > u64_Now) ? (tim_Timepoint - u64_Now) : 0;
}

static inline int RTOS_TimepointCmp(RTOS_Timepoint_t a, RTOS_Timepoint_t b)
{
    if (a == b)
    {
        return 0;
    }
    return a < b ? -1 : 1;
}

#endif

    /* ======================================================================= */
    /* ============================= D E F I N E S =========================== */
    /* ======================================================================= */

    /* ======================================================================= */
    /* ========================== P R O T O T Y P E S ======================== */
    /* ======================================================================= */

#ifdef __cplusplus
}
#endif

#endif /* LION_RTOS_TIMEOUT_H__ */