#include "lion/ICC_Client.h"
#include "lion/log/log.h"
#include "lion/rtos/RTOS_Sleep.h"
#include <errno.h>
#include <malloc.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
LOG_MODULE_DECLARE(icc_client);

static void icc_client_SetState(ICC_Client_Context_t *ptrContext, ICC_Client_State_t enmState)
{
    RTOS_MutexLock(&ptrContext->strStateLock, RTOS_FOREVER);
    ptrContext->enmState = enmState;
    RTOS_MutexUnlock(&ptrContext->strStateLock);
    RTOS_CondVarBroadcast(&ptrContext->strStateChanged);
}

static size_t icc_client_StreamWrite(const uint8_t *ptrData, size_t sizLength, bool bolAll, void *ptrUserData)
{
    ICC_Client_Context_t *ptrContext = ptrUserData;
    ssize_t               sizResult  = send(ptrContext->fdSocket, ptrData, sizLength, MSG_DONTWAIT | MSG_NOSIGNAL);
    if (sizResult < 0)
    {
        sizResult = 0;
        switch (errno)
        {
            case EAGAIN:
                /* Do Nothing */
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                ptrContext->bolBroken = true;
                break;
        }
    }
    return sizResult;
}

static size_t icc_client_StreamRead(uint8_t *ptrData, size_t sizMaxLength, bool bolAll, void *ptrUserData)
{
    ICC_Client_Context_t *ptrContext = ptrUserData;
    ssize_t sizResult = recv(ptrContext->fdSocket, ptrData, sizMaxLength, MSG_DONTWAIT | (bolAll ? MSG_WAITALL : 0));
    if (sizResult < 0)
    {
        sizResult = 0;
        switch (errno)
        {
            case EAGAIN:
                /* Do Nothing */
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                ptrContext->bolBroken = true;
                break;
        }
    }
    return sizResult;
}

static void *icc_client_ThreadEntry(void *arg)
{
    ICC_Client_Context_t *ptrContext = arg;

    while (ptrContext->bolKeepRunning)
    {
        icc_client_SetState(ptrContext, ICC_CLIENT_DISCONNECTED);
        ptrContext->fdSocket = socket(AF_UNIX, SOCK_STREAM, 0);
        if (ptrContext->fdSocket < 0)
        {
            LOG_ERR("Unable to create socket \"%s\"", strerror(errno));
            RTOS_Sleep(RTOS_SECONDS(1));
            continue;
        }

        struct sockaddr_un strSocketAddress;
        strSocketAddress.sun_family = AF_UNIX;
        strcpy(strSocketAddress.sun_path, ptrContext->szSocketPath);

        if (connect(ptrContext->fdSocket, (const struct sockaddr *)&strSocketAddress, sizeof(strSocketAddress)))
        {
            LOG_ERR("Unable to connect socket \"%s\"", strerror(errno));
            RTOS_Sleep(RTOS_SECONDS(1));
            shutdown(ptrContext->fdSocket, SHUT_RDWR);
            continue;
        }

        ptrContext->bolBroken = false;
        ICC_Stream_Init(&ptrContext->strStream, icc_client_StreamWrite, icc_client_StreamRead, ptrContext);
        ICC_AddTransport(ptrContext->ptrICC, &ptrContext->strStream.transport);

        LOG_INF("Connected");
        icc_client_SetState(ptrContext, ICC_CLIENT_CONNECTED);

        while (ptrContext->bolKeepRunning && !ptrContext->bolBroken)
        {
            RTOS_Sleep(RTOS_MSEC(100));
        }
        ICC_RemoveTransport(ptrContext->ptrICC, &ptrContext->strStream.transport);

        LOG_INF("Disconnected");

        shutdown(ptrContext->fdSocket, SHUT_RDWR);
    }

    icc_client_SetState(ptrContext, ICC_CLIENT_DISABLED);

    return NULL;
}

void ICC_Client_Init(ICC_Client_Context_t *ptrContext, ICC_Context_t *ptrICC, const char *szSocketPath)
{
    ptrContext->ptrICC         = ptrICC;
    ptrContext->szSocketPath   = szSocketPath;
    ptrContext->bolKeepRunning = false;
    ptrContext->enmState       = ICC_CLIENT_DISABLED;
    RTOS_MutexInit(&ptrContext->strStateLock);
    RTOS_CondVarInit(&ptrContext->strStateChanged);
}

int ICC_Client_Start(ICC_Client_Context_t *ptrContext)
{
    int intResult = 0;
    if (ptrContext->bolKeepRunning)
    {
        /* Already Running: Do nothing */
        intResult = -EALREADY;
    }
    else
    {
        ptrContext->bolKeepRunning = true;
        pthread_create(&ptrContext->fdThread, NULL, icc_client_ThreadEntry, ptrContext);
    }
    return intResult;
}

int ICC_Client_Stop(ICC_Client_Context_t *ptrContext)
{
    int intResult = 0;
    if (ptrContext->bolKeepRunning)
    {
        ptrContext->bolKeepRunning = false;
        /* TODO: Use pthread_kill() to interrupt any system calls */
        // pthread_kill(ptrContext->fdThread, SIGUSR1);
        pthread_join(ptrContext->fdThread, NULL);
    }
    else
    {
        /* Already Stopped: Do nothing */
        intResult = -EALREADY;
    }
    return intResult;
}

ICC_Client_State_t ICC_Client_GetState(ICC_Client_Context_t *ptrContext)
{
    RTOS_MutexLock(&ptrContext->strStateLock, RTOS_FOREVER);
    ICC_Client_State_t enmState = ptrContext->enmState;
    RTOS_MutexUnlock(&ptrContext->strStateLock);
    return enmState;
}

int ICC_Client_WaitForState(ICC_Client_Context_t *ptrContext, ICC_Client_State_t enmState, RTOS_Timeout_t strTimeout)
{
    int              intRC         = 0;
    RTOS_Timepoint_t tim_Timepoint = RTOS_TimepointCalc(strTimeout);
    RTOS_MutexLock(&ptrContext->strStateLock, RTOS_FOREVER);
    while (ptrContext->enmState != enmState)
    {
        intRC = RTOS_CondVarWait(&ptrContext->strStateChanged, &ptrContext->strStateLock,
                                 RTOS_TimepointToTimeout(tim_Timepoint));
        if (intRC == -EAGAIN)
        {
            break;
        }
    }
    RTOS_MutexUnlock(&ptrContext->strStateLock);
    return intRC;
}